//Micro ST beta version 10
//Author Sergey Bulba, svbulba@gmail.com, http://bulba.untergrund.net
//(c)2001-2017 Bulba S.V.

unit Main;

{$mode objfpc}{$H+}

interface

uses
  LCLIntf, LCLType, SysUtils, Classes, Graphics, Controls,
  Forms, Dialogs, StdCtrls, Buttons, ExtCtrls, ComCtrls,
  digsound, simpleipc;

type

  { TForm1 }

  TForm1 = class(TForm)
   CheckBox2: TCheckBox;
    Edit9: TEdit;
    Label1: TLabel;
    Label12: TLabel;
    Label2: TLabel;
    Label3: TLabel;
    Label4: TLabel;
    Label5: TLabel;
    Label6: TLabel;
    Edit1: TEdit;
    Edit2: TEdit;
    Edit3: TEdit;
    Edit4: TEdit;
    Edit5: TEdit;
    Edit6: TEdit;
    SpeedButton1: TSpeedButton;
    OpenDialog1: TOpenDialog;
    SpeedButton10: TSpeedButton;
    SpeedButton2: TSpeedButton;
    SpeedButton3: TSpeedButton;
    SpeedButton4: TSpeedButton;
    SpeedButton5: TSpeedButton;
    SpeedButton6: TSpeedButton;
    Label7: TLabel;
    Edit7: TEdit;
    SpeedButton7: TSpeedButton;
    SpeedButton8: TSpeedButton;
    ScrollBar1: TScrollBar;
    Label8: TLabel;
    Label9: TLabel;
    Label10: TLabel;
    SpeedButton9: TSpeedButton;
    StatusBar1: TStatusBar;
    Timer1: TTimer;
    Edit8: TEdit;
    Label11: TLabel;
    CheckBox1: TCheckBox;
    procedure CheckBox2Change(Sender: TObject);
    procedure IPCMessage(Sender: TObject);
    procedure FormDropFiles(Sender: TObject; const FileNames: array of String);
    procedure FormShow(Sender: TObject);
    procedure SpeedButton1Click(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure SpeedButton2Click(Sender: TObject);
    procedure ClearInfo;
    procedure LoadSNDHFile(FileName:string);
    procedure LoadMUSFile(FileName:string);
    procedure ExtractTextInfo;
    procedure InitAtariST;
    procedure StartPlaying;
    function FillBuffers:boolean;
    procedure MakeBuffer;
    procedure CheckBuffers;
    procedure SpeedButton3Click(Sender: TObject);
    procedure StopPlaying;
    procedure FormDestroy(Sender: TObject);
    procedure SpeedButton4Click(Sender: TObject);
    procedure SpeedButton6Click(Sender: TObject);
    procedure SpeedButton5Click(Sender: TObject);
    procedure SpeedButton7Click(Sender: TObject);
    procedure PauseSong;
    procedure OpenAndPlay(FileName:string);
    procedure CheckCommandLine;
    procedure CommandLineInterpreter(ParamStrI:string);
    procedure SpeedButton8Click(Sender: TObject);
    procedure ScrollBar1Scroll(Sender: TObject; ScrollCode: TScrollCode;
      var ScrollPos: Integer);
    procedure Timer1Timer(Sender: TObject);
    procedure CheckBox1Click(Sender: TObject);
  private
    { Private declarations }
  public
    { Public declarations }
    TimerPeriod:integer;
    Paused:boolean;
    IsPlaying:boolean;
    FileOpened:boolean;
    ParamFile:string;
    FileEx:boolean;
  end;

procedure ShowError(const s:string);
{$IFDEF Dbg}
procedure DbgStr(const s:string);
{$ENDIF}

var
  Form1: TForm1;
  IPCServer:TSimpleIPCServer;

 BaseSample:int64;
 ShownTime:integer;
 SBMoving:boolean;

implementation

uses
  Params, ay, atari, mc68000, sndh;

{$R *.lfm}

type
 FTypes = (SNDHFile,MUSFile,UnknownFile);
var
 FileType:FTypes;
 NumberOfSongs:integer;
 PlayTimes:TArrayOfInteger;
 SongNames:TArrayOfString;

{$IFDEF Dbg}
procedure DbgStr(const s:string);
var
 f:Text;
 fn:string;
begin
 fn := ExtractFilePath(ParamStr(0))+'_debuglog.txt';
 AssignFile(f,fn);
 if FileExists(fn) then
  Append(f)
 else
  Rewrite(f);
 Writeln(f,s);
 CloseFile(f);
end;
{$ENDIF Dbg}

function TimeSToStr(ms:integer):string;
begin
SetLength(Result,4);
Result[4] := char(ms mod 10 + 48);
ms := ms div 10;
Result[3] := char(ms mod 6 + 48);
ms := ms div 6;
Result[2] := ':';
Result[1] := char(ms mod 10 + 48);
ms := ms div 10;
if ms = 0 then exit;
Result := char(ms mod 6 + 48) + Result;
ms := ms div 6;
if ms = 0 then exit;
Result := IntToStr(ms) + ':' + Result;
end;

procedure TForm1.ExtractTextInfo;
var
 COMM,TITL,RIPP,CONV,YEAR,OtherShortDesc:string;
 PlayGen:TPlayGen; //ignored, always used VBL for the moment
 MuMo:boolean; //ignored, depricated
begin
if not sndh_ExtractTextInfo(SNDHBuffer,COMM,TITL,RIPP,CONV,YEAR,OtherShortDesc,
   NumberOfSongs,CurrentSong,PlayFreq,PlayGen,
   PlayTimes,SongNames,MuMo) then
 exit;
Edit1.Text := COMM;
Edit2.Text := TITL;
Edit3.Text := RIPP;
Edit4.Text := CONV;
Edit8.Text := YEAR;
Edit6.Text := IntToStr(NumberOfSongs);
Edit5.Text := OtherShortDesc;
end;

procedure TForm1.ClearInfo;
begin
Edit1.Clear;
Edit2.Clear;
Edit3.Clear;
Edit4.Clear;
Edit5.Clear;
Edit6.Text := '1';
Edit7.Clear;
Edit8.Clear;
Edit9.Clear;
Label8.Caption:='0:00';
Label10.Caption:='0:00';
NumberOfSongs := 1;
CurrentSong := 1;
PlayFreq := 50;
//PlayGen := GenVBL;
SetLength(PlayTimes,1);
SetLength(SongNames,1);
PlayTimes[0] := 50 * 300; //5 minutes
SongNames[0] := '';
end;

procedure TForm1.LoadSNDHFile(FileName:string);
{$IFDEF Dbg}
var
 f:file;
{$ENDIF Dbg}
begin
ClearInfo;
if not sndh_UnpackFile(FileName,SNDHBuffer) then
 ShowError('Error: damaged file - SNDH ICE depacking');
{$IFDEF Dbg}
AssignFile(f,ExtractFilePath(ParamStr(0))+'_debug.sndh');
Rewrite(f,1);
BlockWrite(f,SNDHBuffer[0],Length(SNDHBuffer));
CloseFile(f);
{$ENDIF Dbg}
ExtractTextInfo;
FileType := SNDHFile;
end;

procedure TForm1.LoadMUSFile(FileName:string);
var
 f:file;
 UnpackedSize:integer;
begin
ClearInfo;
AssignFile(f,FileName);
Reset(f,1);
UnpackedSize := FileSize(f);
SetLength(SNDHBuffer,UnpackedSize);
BlockRead(f,SNDHBuffer[0],UnpackedSize);
CloseFile(f);
FileType := MUSFile;
end;

function TForm1.FillBuffers:boolean;
var
 i,Res:integer;
begin
Result := False;
for i := 0 to 1 do
 begin
  Res := digsound_get_buffer(PBuffer,BufferLength);
  if Res = Ord(digsound_rc_no_buffer) then break;
  if Res = 0 then
   begin
    MakeBuffer;
    if Atari_ExecutionError then break;
    if BuffLen = 0 then
     StopPlaying //todo ???
    else
     begin
      Res := digsound_push(BuffLen);
      if Res > 0 then
       begin
        ShowError(PChar(digsound_geterrortext(Res))+' digsound_push (check)');
        break;
       end;
     end;
    Result := True;
    if Real_End then
     break;
   end
  else if Res > 0 then
   begin
    ShowError(PChar(digsound_geterrortext(Res))+' digsound_get_buffer (check)');
    break;
   end;
 end;
end;

procedure TForm1.InitAtariST;
begin
IntAddr := 8;
if FileType <> SNDHFile then
 IntAddr := 6;
Atari_PrepEmu(SpeedButton10.Down);
TickCountMax := PlayTimes[CurrentSong - 1];
Edit7.Text := IntToStr(CurrentSong);
end;

procedure TForm1.StartPlaying;
var
 Res:integer;
begin
if not FileOpened then exit;
if IsPlaying then exit;

BaseSample := 0;
Seeking := False;
ShownTime := 0;
Label8.Caption := '0:00';
Label10.Caption := TimeSToStr(round(PlayTimes[CurrentSong - 1] / PlayFreq));
Edit9.Text := SongNames[CurrentSong - 1];
SBMoving := False;
ScrollBar1.Max := PlayTimes[CurrentSong - 1];
ScrollBar1.Position := 0;
ScrollBar1.SmallChange:=ScrollBar1.Max div 100;
ScrollBar1.LargeChange:=ScrollBar1.Max div 10;
ShowError('');

InitAtariST;

{$ifdef DebugI}
ShowMessage('Open Console to work in interactive MC68000 debugger (Ctrl+Alt+O in Lazarus) or restart from Console.'+#13+
 'In Console type ? to help or q to quit.');
while cpudebug_interactive(1,nil,nil,nil,nil) <> -1 do ;
{$endif}

Res := digsound_open({-1,}2,44100,16,2,370);

if Res <> 0 then
 ShowError(PChar(digsound_geterrortext(Res))+' digsound_open')
else if FillBuffers then
 begin
  IsPlaying := True;
  Paused := False;
 end
else
 begin
  Res := digsound_close;
  if Res <> 0 then
   ShowError(PChar(digsound_geterrortext(Res))+' digsound_close(init1)');
 end;
if IsPlaying then
 begin
  if not SpeedButton9.Down then
   SpeedButton9.Enabled := False;
  if not SpeedButton10.Down then
   SpeedButton10.Enabled := False;
 end
else
 Atari_StopEmu;
end;

procedure TForm1.SpeedButton1Click(Sender: TObject);
begin
if OpenDialog1.Execute then
 begin
  OpenDialog1.InitialDir := ExtractFileDir(OpenDialog1.FileName);
  OpenAndPlay(OpenDialog1.FileName);
 end
end;

procedure TForm1.FormCreate(Sender: TObject);
begin
Mono := CheckBox2.Checked;

//TOS Malloc tests:
{InitSTMem(1000,11000);
GetSTMemMax;
GetSTMem(1000);
GetSTMem(500);
FreeSTMem(2000);
FreeSTMem(2000);
FreeSTMem(2001);
GetSTMem(2000);
GetSTMem(3000);
FreeSTMem(2000);
GetSTMem(1000);
FreeSTMem(1000);
FreeSTMem(2000);
GetSTMemMax;
FreeSTMem;}

AtariErrorHandler := @ShowError;

Do_Loop := True;
CheckBox1.Checked := Do_Loop;

OpenDialog1.InitialDir := ExtractFileDir(ParamStr(0));
SetCurrentDir(OpenDialog1.InitialDir);

IsPlaying := False;
FileOpened := False;
FileType := UnknownFile;

Atari_SetDefault;

CalcFiltKoefs;
IPCServer.OnMessage := @IPCMessage;
end;

procedure TForm1.SpeedButton2Click(Sender: TObject);
begin
Application.MessageBox(
 'Micro ST beta version 10'#13 +
 'Author Sergey Bulba'#13+
 'Visit http://bulba.untergrund.net/'#13+
 'Compiled at 19 of November 2017'#13+
 '(c)2001-2017 S.V. Bulba','About program',mb_OK);
end;

{procedure SaveBuffer;
var
 f:file;
begin
AssignFile(f,ExtractFilePath(ParamStr(0))+'test.nowh');
if FileExists(ExtractFilePath(ParamStr(0))+'test.nowh') then
 Reset(f,1)
else
 Rewrite(f,1);
Seek(f,FileSize(f));
BlockWrite(f,PBuffer^,BuffLen);
CloseFile(f);
end;}

procedure TForm1.MakeBuffer;
begin
BuffLen := 0;
if IntFlag then
 begin
  IntFlag := False;
  Synthesizer_Stereo16;
  if IntFlag then exit;
 end;
Atari_CheckOuts; //todo: вывести пропущенные данные с точностью до такта (для тестов sndh46lf\Tao\Songs_That_Make_You_Go_Mmh\Commando_Hiscore_digi.sndh)
//if Real_End then exit;
while (BuffLen < BufferLength) and not Real_End do //проверить, по Real_End всегда доиграется последний тик или нет?
 begin
  Atari_Emulate;
  if Atari_ExecutionError then
   begin
    StopPlaying;
    break;
   end;
  if (BuffLen < BufferLength) and ((s68000readOdometer >= 150000) or Real_End) then
   SynthesizerSNDH;
end;
end;

procedure TForm1.CheckBuffers;
begin
if Atari_ExecutionError then exit;
if not IsPlaying then exit;

if Real_End then
 begin
  if digsound_check = Ord(digsound_rc_done) then
   StopPlaying;
  exit;
 end;

FillBuffers;
end;

procedure TForm1.SpeedButton3Click(Sender: TObject);
begin
StopPlaying;
end;

procedure TForm1.StopPlaying;
var
  Res:integer;
begin
if IsPlaying then
 begin
  IsPlaying := False;
  SpeedButton9.Enabled := True;
  SpeedButton10.Enabled := True;
  Res := digsound_close;
  if Res <> 0 then
   ShowError(PChar(digsound_geterrortext(Res))+' digsound_close');
  Atari_StopEmu;
 end;
end;

procedure TForm1.FormDestroy(Sender: TObject);
begin
IPCServer.OnMessage := nil;
StopPlaying;
Atari_Free;
end;

procedure TForm1.SpeedButton4Click(Sender: TObject);
begin
StartPlaying;
end;

procedure TForm1.SpeedButton6Click(Sender: TObject);
begin
if CurrentSong < NumberOfSongs then
 begin
  StopPlaying;
  inc(CurrentSong);
  StartPlaying;
 end;
end;

procedure TForm1.SpeedButton5Click(Sender: TObject);
begin
if CurrentSong > 1 then
 begin
  StopPlaying;
  dec(CurrentSong);
  StartPlaying;
 end;
end;

procedure TForm1.PauseSong;
var
  Res:integer;
begin
if not IsPlaying then
 exit;
if Paused then
 begin
  Res := digsound_continue;
  if Res <> 0 then
   ShowError(PChar(digsound_geterrortext(Res))+' digsound_continue')
  else
   Paused := False;
 end
else
 begin
  Res := digsound_pause;
  if Res <> 0 then
   ShowError(PChar(digsound_geterrortext(Res))+' digsound_pause')
  else
   Paused := True;
 end;
end;

procedure TForm1.SpeedButton7Click(Sender: TObject);
begin
PauseSong;
end;

procedure TForm1.IPCMessage(Sender: TObject);
begin
CommandLineInterpreter(IPCServer.StringMessage);
end;

procedure TForm1.CheckBox2Change(Sender: TObject);
begin
Mono := CheckBox2.Checked;
Calculate_Level_Tables;
end;

procedure TForm1.FormDropFiles(Sender: TObject; const FileNames: array of String);
begin
if Length(FileNames) = 0 then exit;
StopPlaying;
OpenAndPlay(FileNames[0]);
end;

procedure TForm1.FormShow(Sender: TObject);
begin
{$ifdef Windows}
  ScrollBar1.SetFocus; //GTK2 LCL bug? Or feature?
{$else}
  CheckBox1.SetFocus;
{$endif}
end;

procedure TForm1.OpenAndPlay(FileName:string);
var
 Extn:string;
begin
ShowError('');
Extn := LowerCase(ExtractFileExt(FileName));
if (Extn <> '.snd') and (Extn <> '.sndh') and (Extn <> '.mus') then
 begin
  ShowError('Not supported type: ' + Extn + ' - Error');
  exit;
 end;
StopPlaying;
{$IFDEF Dbg}
DbgStr(FileName);
{$ENDIF Dbg}
if (Extn = '.snd') or (Extn = '.sndh') then
 LoadSNDHFile(FileName)
else
 LoadMUSFile(FileName);
FileOpened := True;
StartPlaying;
end;

procedure TForm1.CheckCommandLine;
var
 i:integer;
begin
if ParamCount <> 0 then
 begin
  CommandLineInterpreter('?');
  for i:=1 to ParamCount do
   CommandLineInterpreter(ParamStr(i));
  CommandLineInterpreter('');
 end;
end;

procedure TForm1.CommandLineInterpreter(ParamStrI:string);
var
 LKey:char;
{ j:longword;
 NumStr:string;
 ErrPos,NewFrq:integer;}
begin
if ParamStrI = '?' then //start marker
 FileEx := False
else if ParamStrI <> '' then //end marker
 begin
  if (Length(ParamStrI) > 1) and
     ({$ifdef Windows}(ParamStrI[1] = '/') or {$endif}(ParamStrI[1] = '-')) then
   begin
    LKey := CharLower(ParamStrI[2]);
    case LKey of
    'v': Application.BringToFront;
    {'r': begin
         if StrLen(HAddr)=3 then
          begin
           NumStr := char(CharLower(pchar(HAddr[2])));
           if (NumStr = 'n') or (NumStr = 'h') then
            begin
             if NumStr = 'n' then
              SetPriority(NORMAL_PRIORITY_CLASS)
             else if NumStr='h' then
              SetPriority(HIGH_PRIORITY_CLASS);
             Case Priority of
              NORMAL_PRIORITY_CLASS:Form2.RadioButton1.Checked := True;
              HIGH_PRIORITY_CLASS:  Form2.RadioButton2.Checked := True;
             end;
            end;
          end;
         end;
    'h': begin
         j := 3; NumStr := '';
         while (j <= StrLen(HAddr)) and (j < 34) do
          begin
           NumStr := NumStr + HAddr[j-1]; inc(j);
          end;
         NumStr := AnsiUpperCase(NumStr);
         if NumStr = 'AYMONO' then NewFrq := 0 else
         if NumStr = 'YMMONO' then NewFrq := 1 else
         if NumStr = 'AYABC' then NewFrq := 2 else
         if NumStr = 'YMABC' then NewFrq := 3 else
         if NumStr = 'AYACB' then NewFrq := 4 else
         if NumStr = 'YMACB' then NewFrq := 5 else
         if NumStr = 'AYBAC' then NewFrq := 6 else
         if NumStr = 'YMBAC' then NewFrq := 7 else
         if NumStr = 'AYBCA' then NewFrq := 8 else
         if NumStr = 'YMBCA' then NewFrq := 9 else
         if NumStr = 'AYCAB' then NewFrq := 10 else
         if NumStr = 'YMCAB' then NewFrq := 11 else
         if NumStr = 'AYCBA' then NewFrq := 12 else
         if NumStr = 'YMCBA' then NewFrq := 13 else NewFrq := -1;
         if NewFrq >= 0 then
          begin
           Form2.ComboBox1.ItemIndex := NewFrq;
           SetChipAndChannels(NewFrq);
          end;
         end;
    'u': begin
         j := 3; NumStr := '';
         while (j <= StrLen(HAddr)) and (j < 34) do
          begin
           NumStr := NumStr + HAddr[j-1];
           inc(j);
          end;
         Val(NumStr,NewFrq,ErrPos);
         if (ErrPos = 0) and (NewFrq >= 0) and (NewFrq<= 32768) then
          begin
           Form2.Edit1.Text := IntToStr(NewFrq);
           BeeperMax := NewFrq;
           BeeperLevel := round(-BeeperMax*(exp(CurrVolume*ln(2)/
                        (MZVolume.zy2 - MZVolume.zy1)) - 1))
          end;
         end;
    'p': begin
         j := 3; NumStr := '';
         while j <= StrLen(HAddr) do
          begin
           NumStr := NumStr + HAddr[j - 1];
           inc(j)
          end;
         if NumStr = '' then
          j := 0
         else
          j := 1;
         if LoadSkin(j,NumStr) then
          begin
           Form2.Edit2.Text := NumStr;
           PrintNumber(1);
           PrintNumber(2);
           Repaint
          end
         end;}
    end;
   end
  else if FileExists(ParamStrI) then
   begin
    FileEx := True;
    StopPlaying;
    ParamFile := ParamStrI;
   end
 end
else
 begin
  if FileEx then
   begin
    StopPlaying;
    OpenAndPlay(ParamFile);
    SetLength(ParamFile,0);
   end;
 end;
end;

procedure TForm1.SpeedButton8Click(Sender: TObject);
begin
if Form2.ShowModal = mrOK then
 begin
  MainClockFreq := Form2.MainClc;
  MC68000Freq := Form2.MCClc;
  MFPFreq := Form2.MFPClc;
  AyFreq := Form2.YMClc;
  VBLFreq := Form2.VBClc;
  VBLPeriod := round (MC68000Freq / VBLFreq);
  FrqAyByFrqMC68000 := round(AyFreq/MC68000Freq/8*4294967296);
  Delay_In_Tiks := round(8192/44100*AyFreq);
  MCbyMFP := MC68000Freq/MFPFreq;
  IsFilt := Form2.CheckBox1.Checked;
  CalcFiltKoefs;
 end;
end;

procedure SeekTo(TargetTick:integer);
begin
 if not Form1.IsPlaying then exit;
 if Form1.Paused then exit;
 BuffLen := 0;
 if not Atari_SeekTo(TargetTick) then
  begin
   Form1.StopPlaying;
   exit;
  end;
 BaseSample := trunc(TickCount / VBLFreq * 44100 + 0.5);
 if IsFilt then
  begin
   FillChar(Filt_XL[0],(Filt_M + 1) * 4,0);
   FillChar(Filt_XR[0],(Filt_M + 1) * 4,0);
   Filt_I := 0;
  end;
 digsound_reset;
end;

procedure TForm1.ScrollBar1Scroll(Sender: TObject; ScrollCode: TScrollCode;
  var ScrollPos: Integer);
begin
if not IsPlaying or Paused then exit;
if ScrollCode <> scEndScroll then
 begin
  SBMoving := True;
  Form1.Label8.Caption := TimeSToStr(trunc(ScrollPos / VBLFreq + 0.5));
  Form1.Label8.Repaint;
 end
else
 begin
  SBMoving := False;
  SeekTo(ScrollPos);
 end;
end;

procedure TForm1.Timer1Timer(Sender: TObject);
var
 s:int64;
 t:double;
 i:integer;
begin
if not IsPlaying or Paused then exit;
if not Seeking and not SBMoving then
 if digsound_getposition(s{%H-}) = 0 then
   begin
    t := (BaseSample + s) / 44100;
    Form1.ScrollBar1.Position := round(t * VBLFreq);
    i := round(t);
    if i <> ShownTime then
     begin
      ShownTime := i;
      Form1.Label8.Caption := TimeSToStr(i);
      Form1.Label8.Repaint;
     end;
   end;
CheckBuffers;
end;

procedure TForm1.CheckBox1Click(Sender: TObject);
begin
Do_Loop := CheckBox1.Checked;
end;

procedure ShowError(const s:string);
begin
{$IFDEF Dbg}
if (s <> '') and (s <> Form1.StatusBar1.SimpleText) then
 DbgStr(#9+s);
{$ENDIF Dbg}
Form1.StatusBar1.SimpleText:=s;
Form1.StatusBar1.Hint:=s;
Form1.StatusBar1.ShowHint:=s <> '';
end;

end.

