unit ay;

{$mode objfpc}{$H+}
{$asmmode intel}

interface

uses
 Classes, SysUtils;

procedure SetAYRegister(Num:integer;Value:byte);
procedure SetEnvelopeRegister(Value:byte);
procedure SetMixerRegister(Value:byte);
procedure SetAmplA(Value:byte);
procedure SetAmplB(Value:byte);
procedure SetAmplC(Value:byte);
procedure Synthesizer_Stereo16;
procedure InitAYEmu;
procedure Calculate_Level_Tables;
procedure CalcFiltKoefs;

type
  TArrayOfByte = array of byte;
  TArrayOfInteger = array of integer;
  TArrayOfString = array of string;
  Buf = packed array[0..0] of record
   Left:smallint;
   Right:smallint;
  end;
  PBuf = ^Buf;

  TRegisterAY = packed record
  case Integer of
   0:(Index:array[0..15]of byte);
   1:(TonA,TonB,TonC: word;
      Noise:byte;
      Mixer:byte;
      AmplitudeA,AmplitudeB,AmplitudeC:byte;
      Envelope:word;
      EnvType:byte);
  end;

const
 Filt_M = 32; //powers of 2
type
 TFilt_K = array[0..Filt_M] of integer;

var
  Mono:boolean;
  IsFilt:boolean = True;
  Filt_K,Filt_XL,Filt_XR:TFilt_K;
  Filt_I:integer;
  AyFreq:real;
  PlayFreq:integer;
  TickCount,TickCountMax:integer;
  Delay_In_Tiks:dword;
  Real_End:boolean;
  Do_Loop:boolean;
  Seeking:boolean;
  RegisterAY:TRegisterAY;
  Current_RegisterAY:byte;
  First_Period:boolean;
  Ton_Counter_A,
  Ton_Counter_B,
  Ton_Counter_C,
  Noise_Counter:packed record
   case integer of
    0:(Lo:word;
       Hi:word);
    1:(Re:longword);
   end;
  Envelope_Counter:packed record
   case integer of
   0:(Lo:dword;
      Hi:dword);
   1:(Re:int64);
   end;
  Ton_A,Ton_B,Ton_C:byte;
  Noise:packed record
   case boolean of
    True: (Seed:longword);
    False:(Low:word;
           Val:word);
   end;
  Ampl:smallint;
  Case_EnvType:procedure;
  Ton_EnA,Ton_EnB,
  Ton_EnC,Noise_EnA,
  Noise_EnB,Noise_EnC:boolean;
  Envelope_EnA,Envelope_EnB,Envelope_EnC:boolean;
  Level_AR,Level_AL,
  Level_BR,Level_BL,
  Level_CR,Level_CL:array[0..31]of Integer;
  Left_Chan,Right_Chan:integer;
  Tick_Counter:byte;
  Tik:packed record
   Case Integer of
    0:(Lo:Word;
       Hi:word);
    1:(Re:dword);
   end;
  Current_Tik:longword;
  BuffLen,BufferLength:integer;
  PBuffer:PBuf;
  Number_Of_Tiks:packed record
   case boolean of
    false:(lo:longword;
           hi:longword);
    true: (re:int64);
   end;
  IntFlag:boolean;
  IntDMASnd:boolean;

implementation

uses
 atari;

const
 { (c)Hacker KAY }
 Amplitudes_YM:array[0..31]of Word=
   (0, 0, $F8, $1C2, $29E, $33A, $3F2, $4D7, $610, $77F, $90A, $A42,
   $C3B, $EC2, $1137, $13A7, $1750, $1BF9, $20DF, $2596, $2C9D, $3579,
   $3E55, $4768, $54FF, $6624, $773B, $883F, $A1DA, $C0FC, $E094, $FFFF);

//sorry for assembler, I can't make effective qword procedure on pascal...
function ApplyFilter(Lev:integer;var Filt_X:TFilt_K):integer;
{$ifdef cpu32}
begin
asm
        push    ebx
        push    esi
        push    edi
        add     esp,-8
//        mov     ecx,Filt_M

//        mov     edi,Filt_K
        lea     edi,Filt_K

//        lea     esi,edi+ecx*4
        lea     esi,edi[Filt_M*4]

//        mov     ebx,[edx]
        mov     ebx,edx

        mov     ecx,Filt_I
        mov     [ebx+ecx*4],eax
        imul    dword ptr [edi]
        mov     [esp],eax
        mov     [esp+4],edx
@lp:    dec     ecx
        jns     @gz
        mov     ecx,Filt_M
@gz:    mov     eax,[ebx+ecx*4]
        add     edi,4
        imul    dword ptr [edi]
        add     [esp],eax
        adc     [esp+4],edx
        cmp     edi,esi
        jnz     @lp
        mov     Filt_I,ecx
        pop     eax
        pop     edx
        pop     edi
        pop     esi
        pop     ebx
        test    edx,edx
        jns     @nm
        add     eax,0FFFFFFh
        adc     edx,0
@nm:    shrd    eax,edx,24
end;
{$else}
var
 Res:int64;
 j:integer;
begin
Filt_X[Filt_I] := Lev;
Res := Lev*Filt_K[0];
for j := 1 to Filt_M do
 begin
  if Filt_I > 0 then
   Dec(Filt_I)
  else
   Filt_I := Filt_M;
  Inc(Res,Filt_X[Filt_I]*Filt_K[j]);
 end;
Result := Res div $1000000;
{$endif}
end;

procedure Synthesizer_Logic_Q;
begin
inc(Ton_Counter_A.Hi);
if Ton_Counter_A.Hi >= RegisterAY.TonA then
 begin
  Ton_Counter_A.Hi := 0;
  Ton_A := Ton_A xor 1;
 end;
inc(Ton_Counter_B.Hi);
if Ton_Counter_B.Hi >= RegisterAY.TonB then
 begin
  Ton_Counter_B.Hi := 0;
  Ton_B := Ton_B xor 1;
 end;
inc(Ton_Counter_C.Hi);
if Ton_Counter_C.Hi >= RegisterAY.TonC then
 begin
  Ton_Counter_C.Hi := 0;
  Ton_C := Ton_C xor 1;
 end;
inc(Noise_Counter.Hi);
if (Noise_Counter.Hi and 1 = 0) and
   (Noise_Counter.Hi >= RegisterAY.Noise shl 1) then
 begin
  Noise_Counter.Hi := 0;
  {$ifdef cpu32}
  asm
  mov eax,Noise.Seed
  shld edx,eax,16
  shld ecx,eax,19
  xor ecx,edx
  and ecx,1
  add eax,eax
  and eax,$1ffff
  inc eax
  xor eax,ecx
  mov Noise.Seed,eax
  end;
  {$else}
  with Noise do
   Seed := (((Seed shl 1) or 1) xor ((Seed shr 16) xor (Seed shr 13) and 1)) and $1ffff;
  {$endif}
 end;
if Envelope_Counter.Hi = 0 then Case_EnvType;
inc(Envelope_Counter.Hi);
if Envelope_Counter.Hi >= RegisterAY.Envelope then
 Envelope_Counter.Hi := 0;
end;

procedure Synthesizer_Mixer_Q;
var
 LevL,LevR,k:integer;
begin
LevL := 0;
LevR := LevL;

Atari_MixDMASnd(LevL,LevR);

k := 1;
if Ton_EnA then k := Ton_A;
if Noise_EnA then k := k and Noise.Val;
if k <> 0 then
 begin
  if Envelope_EnA then
   begin
    inc(LevL,Level_AL[RegisterAY.AmplitudeA * 2 + 1]);
    inc(LevR,Level_AR[RegisterAY.AmplitudeA * 2 + 1])
   end
  else
   begin
    inc(LevL,Level_AL[Ampl]);
    inc(LevR,Level_AR[Ampl])
   end
 end;

k := 1;
if Ton_EnB then k := Ton_B;
if Noise_EnB then k := k and Noise.Val;
if k <> 0 then
 if Envelope_EnB then
  begin
   inc(LevL,Level_BL[RegisterAY.AmplitudeB * 2 + 1]);
   inc(LevR,Level_BR[RegisterAY.AmplitudeB * 2 + 1])
  end
 else
  begin
   inc(LevL,Level_BL[Ampl]);
   inc(LevR,Level_BR[Ampl])
  end;

k := 1;
if Ton_EnC then k := Ton_C;
if Noise_EnC then k := k and Noise.Val;
if k <> 0 then
 if Envelope_EnC then
  begin
   inc(LevL,Level_CL[RegisterAY.AmplitudeC * 2 + 1]);
   inc(LevR,Level_CR[RegisterAY.AmplitudeC * 2 + 1]);
  end
 else
  begin
   inc(LevL,Level_CL[Ampl]);
   inc(LevR,Level_CR[Ampl]);
  end;

if IsFilt then
 begin
  k := Filt_I;
  LevL := ApplyFilter(LevL,Filt_XL);
  Filt_I := k;
  LevR := ApplyFilter(LevR,Filt_XR);
 end;

inc(Left_Chan,LevL);
inc(Right_Chan,LevR);
end;

procedure Synthesizer_Stereo16;
var
 Tmp:integer;
begin
repeat
Synthesizer_Logic_Q;
Synthesizer_Mixer_Q;
Inc(Current_Tik);
Inc(Tick_Counter);
if Tick_Counter >= Tik.Hi then
 begin
  Inc(Tik.Re,Delay_In_Tiks);
  Dec(Tik.Hi,Tick_Counter);
  Tmp := Left_Chan div Tick_Counter;
  if Tmp > 32767 then
   Tmp := 32767
  else if Tmp < -32768 then
   Tmp := -32768;
  PBuffer^[BuffLen].Left := Tmp;
  Tmp := Right_Chan div Tick_Counter;
  if Tmp > 32767 then
   Tmp := 32767
  else if Tmp < -32768 then
   Tmp := -32768;
  PBuffer^[BuffLen].Right := Tmp;
  Inc(BuffLen);
  Tmp := 0;
  Left_Chan:= Tmp;
  Right_Chan := Tmp;
  Tick_Counter := Tmp;
  if BuffLen = BufferLength then
   begin
    if Current_Tik < Number_Of_Tiks.Hi then
     IntFlag := True;
    exit;
   end
 end
until Current_Tik >= Number_Of_Tiks.Hi;
Tmp := 0;
Number_Of_Tiks.hi := Tmp;
Current_Tik := Tmp;
end;

procedure Case_EnvType_0_3__9;
begin
if First_Period then
 begin
  dec(Ampl);
  if Ampl = 0 then First_Period := False;
 end;
end;

procedure Case_EnvType_4_7__15;
begin
if First_Period then
 begin
  Inc(Ampl);
  if Ampl = 32 then
   begin
    First_Period := False;
    Ampl := 0;
   end;
 end;
end;

procedure Case_EnvType_8;
begin
Ampl := (Ampl - 1) and 31;
end;

procedure Case_EnvType_10;
begin
if First_Period then
 begin
  dec(Ampl);
  if Ampl < 0 then
   begin
    First_Period := False;
    Ampl := 0;
   end;
 end
else
 begin
  inc(Ampl);
  if Ampl = 32 then
   begin
    First_Period := True;
    Ampl := 31;
   end;
 end;
end;

procedure Case_EnvType_11;
begin
if First_Period then
 begin
  dec(Ampl);
  if Ampl < 0 then
   begin
    First_Period := False;
    Ampl := 31;
   end;
 end;
end;

procedure Case_EnvType_12;
begin
Ampl := (Ampl + 1) and 31;
end;

procedure Case_EnvType_13;
begin
if First_Period then
 begin
  inc(Ampl);
  if Ampl = 32 then
   begin
    First_Period := False;
    Ampl := 31;
   end;
 end;
end;

procedure Case_EnvType_14;
begin
if not First_Period then
 begin
  dec(Ampl);
  if Ampl < 0 then
   begin
    First_Period := True;
    Ampl := 0;
   end;
 end
else
 begin
  inc(Ampl);
  if Ampl = 32 then
   begin
    First_Period := False;
    Ampl := 31;
   end;
 end;
end;

procedure SetMixerRegister(Value:byte);
begin
RegisterAY.Mixer := Value;
Ton_EnA := (Value and 1) = 0;
Noise_EnA := (Value and 8) = 0;
Ton_EnB := (Value and 2) = 0;
Noise_EnB := (Value and 16) = 0;
Ton_EnC := (Value and 4) = 0;
Noise_EnC := (Value and 32) = 0;
end;

procedure SetEnvelopeRegister(Value:byte);
begin
Envelope_Counter.Hi := 0;
First_Period := True;
if (Value and 4) = 0 then
 ampl := 32
else
 ampl := -1;
RegisterAY.EnvType := Value;
case Value of
0..3,9: Case_EnvType := @Case_EnvType_0_3__9;
4..7,15:Case_EnvType := @Case_EnvType_4_7__15;
8:      Case_EnvType := @Case_EnvType_8;
10:     Case_EnvType := @Case_EnvType_10;
11:     Case_EnvType := @Case_EnvType_11;
12:     Case_EnvType := @Case_EnvType_12;
13:     Case_EnvType := @Case_EnvType_13;
14:     Case_EnvType := @Case_EnvType_14;
end;
end;

procedure SetAmplA(Value:byte);
begin
RegisterAY.AmplitudeA := Value;
Envelope_EnA := (Value and 16) = 0;
end;

procedure SetAmplB(Value:byte);
begin
RegisterAY.AmplitudeB := Value;
Envelope_EnB := (Value and 16) = 0;
end;

procedure SetAmplC(Value:byte);
begin
RegisterAY.AmplitudeC := Value;
Envelope_EnC := (Value and 16) = 0;
end;

procedure SetAYRegister(Num:integer;Value:byte);
begin
case Num of
13:
 SetEnvelopeRegister(Value and 15);
1,3,5:
 RegisterAY.Index[Num] := Value and 15;
6:
 RegisterAY.Noise := Value and 31;
7: SetMixerRegister(Value and 63);
8: SetAmplA(Value and 31);
9: SetAmplB(Value and 31);
10:SetAmplC(Value and 31);
0,2,4,11,12,14,15:
 RegisterAY.Index[Num] := Value;
end;
end;

procedure Calculate_Level_Tables;
var
 i,b,l,r:integer;
 Index_A,Index_B,Index_C:integer;
 k:real;
 LC,LE:integer;
begin
Index_A := 255;Index_B:=170;Index_C:=13;
l:=Index_A+Index_B+Index_C;
r:=32767;
l:=255*r div l;
if Mono then
 begin
  LC := 255;
  LE := 255;
 end
else
begin
 LC := 170;
 LE := 13;
end;
for i:=0 to 31 do
 begin
  b:=round(255/255*Amplitudes_YM[i]);
  Level_AL[i]:=round(b/65535*l);
  b:=round(LE/255*Amplitudes_YM[i]);
  Level_AR[i]:=round(b/65535*l);
  b:=round(LC/255*Amplitudes_YM[i]);
  Level_BL[i]:=round(b/65535*l);
  b:=round(LC/255*Amplitudes_YM[i]);
  Level_BR[i]:=round(b/65535*l);
  b:=round(LE/255*Amplitudes_YM[i]);
  Level_CL[i]:=round(b/65535*l);
  b:=round(255/255*Amplitudes_YM[i]);
  Level_CR[i]:=round(b/65535*l);
 end;
k:=(exp(1*ln(2)/1)-1);
for i:=0 to 31 do
 begin
   Level_AL[i]:=round(Level_AL[i]*k);
   Level_AR[i]:=round(Level_AR[i]*k);
   Level_BL[i]:=round(Level_BL[i]*k);
   Level_BR[i]:=round(Level_BR[i]*k);
   Level_CL[i]:=round(Level_CL[i]*k);
   Level_CR[i]:=round(Level_CR[i]*k);
 end;
end;

procedure InitAYEmu;
var
 l:integer;
begin
for l := 0 to 15 do RegisterAY.Index[l] := 0;
SetEnvelopeRegister(0);
SetMixerRegister(0);
SetAmplA(0);
SetAmplB(0);
SetAmplC(0);
First_Period := False;
Ampl := 0;
IntFlag := False;
Number_Of_Tiks.Re := 0;
Current_Tik := 0;
Envelope_Counter.Re:=0;
Ton_Counter_A.Re:=0;
Ton_Counter_B.Re:=0;
Ton_Counter_C.Re:=0;
Noise_Counter.Re:=0;
Ton_A:=0;
Ton_B:=0;
Ton_C:=0;
Left_Chan:=0;Right_Chan:=0;Tick_Counter:=0;
Tik.Re:=Delay_In_Tiks;
Noise.Seed := $ffff;
Current_RegisterAY := 0;
if IsFilt then
 begin
  FillChar(Filt_XL[0],(Filt_M + 1) * 4,0);
  FillChar(Filt_XR[0],(Filt_M + 1) * 4,0);
  Filt_I := 0
 end;
Calculate_Level_Tables;
end;

procedure CalcFiltKoefs;
var
 i,i2,Filt_M2:integer;
 K,F,C:double;
 FKt:array of double;
begin
C := Pi * 44100 / (AyFreq / 8);
SetLength(FKt,Filt_M + 1);
Filt_M2 := Filt_M div 2;
K := 0;
for i := 0 to Filt_M do
 begin
  i2 := i - Filt_M2;
  if i2 = 0 then
   F := C
  else
   F := sin(C * i2) / i2 * (0.54 - 0.46 * cos(2 * Pi / Filt_M * i));
  FKt[i] := F;
  K := K + F;
 end;
for i := 0 to Filt_M do
 Filt_K[i] := round(FKt[i] / K * $1000000);
end;

end.

